/*
 * Copyright (c) 2008
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2008\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: aclc.c 542 2008-12-04 03:30:01Z leres $ (LBL)";
#endif

/*
 * aclc - broccoli acld client
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <broccoli.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>

#include "broccoli.h"
#include "gnuc.h"
#include "version.h"

/* Globals */
int debug;
int verbose;
const char *prog;
const char *addr = "127.0.0.1";
const char *port = "47777";
const char *requestevstr = EV_REQUEST_STR;
const char *replyevstr = EV_REPLY_STR;

const char *file = NULL;
static int timeout = 10;			/* select timeout */
u_int32_t cookie;
int success;

/* Externs */
extern int optind;
extern int opterr;
extern char *optarg;

/* Forwards */
BroEvent *acldevent(const char *);
void callback(BroConn *, void *, BroRecord *);
int command(int, BroConn *, int, char **);
int loop(int, BroConn *);
int main(int, char **);
__dead void usage(void) __attribute__((noreturn));

/* XXX */
int extractbool(BroRecord *, const char *, int *);
int extractstr(BroRecord *, const char *, char *, size_t);
int extracttime(BroRecord *, const char *, double *);
int extractuint32(BroRecord *, const char *, uint32 *);
void *new(size_t, size_t, const char *);

int
main(int argc, char **argv)
{
	int i, op, s, ret;
	char *cp, *ep;
	BroConn *bc;
	struct sockaddr_in sin;

	if (argv[0] == NULL)
		prog = "acld";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	opterr = 0;
	while ((op = getopt(argc, argv, "dDva:c:e:f:P:")) != EOF)
		switch (op) {

		case 'a':
			addr = optarg;
			break;

		case 'd':
			++debug;
			break;

		case 'c':
			// config = optarg;
			break;

		case 'D':
			++bro_debug_messages;
			break;

		case 'e':
			requestevstr = optarg;
			break;

		case 'E':
			replyevstr = optarg;
			break;

		case 'f':
			file = optarg;
			break;

		case 'P':
			port = optarg;
			break;

		case 'v':
			++verbose;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	memset(&sin, 0, sizeof(sin));
#ifdef HAVE_SOCKADDR_SA_LEN
	sin.sin_len = sizeof(sin);
#endif
	sin.sin_family = AF_INET;

	i = strtol(port, &ep, 10);
	if (*ep != '\0' || i < 1 || i > 0xffff) {
		fprintf(stderr, "%s: bad port value (%s)\n", prog, port);
		exit(EX_USAGE);
	}
	sin.sin_port = htons(i);

	/* "*" -> 0.0.0.0 else decode string address */
	if (strcmp(addr, "*") != 0 &&
	    inet_pton(sin.sin_family, addr, &sin.sin_addr) < 0) {
		fprintf(stderr, "%s: bad address value (%s): %s\n",
		    prog, addr, strerror(errno));
		exit(EX_USAGE);
	}

	s = socket(PF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		fprintf(stderr, "%s: socket: %s\n", prog, strerror(errno));
		exit(EX_OSERR);
	}

	while (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
		fprintf(stderr, "%s: connect: %s\n", prog, strerror(errno));
		sleep(1);
	}

	bc = bro_conn_new_socket(s, BRO_CFLAG_RECONNECT);
	if (bc == NULL) {
		fprintf(stderr,
		    "%s: bro_conn_new_socket to %s:%d failed\n",
		    prog, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
		exit(EX_UNAVAILABLE);
	}
	fprintf(stderr, "%s: Connecting to %s:%d\n",
	    prog, inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));

	/* Set the connection class */
	bro_conn_set_class(bc, "acld");

	bro_event_registry_add(bc, replyevstr, (BroEventFunc)callback, NULL);

	while (!bro_conn_connect(bc)) {
		fprintf(stderr, "%s: Failed to connect; retrying...\n", prog);
		sleep(1);
	}

	if (verbose)
		fprintf(stderr, "%s: connected\n", prog);

	ret = 0;
	if (argc != optind)
		ret = command(s, bc, argc - optind, argv + optind);
	else
		ret = loop(s, bc);
	bro_conn_delete(bc);

	exit(ret);
}

BroEvent *
acldevent(const char *buf)
{
	BroEvent *ev;
	BroRecord *rec;
	BroString str;

	rec = bro_record_new();
	if (rec == NULL) {
		fprintf(stderr,
		    "%s: bro_record_new() failed\n",
		    prog);
		return (NULL);
	}

	/* "cmd" */
	bro_string_init(&str);
	bro_string_set(&str, buf);
	if (!bro_record_add_val(rec, EV_CMD, BRO_TYPE_STRING, NULL, &str)) {
		fprintf(stderr, "%s: bro_record_add_val(%s) failed\n",
		    prog, EV_CMD);
		bro_record_free(rec);
		return (NULL);
	}
	bro_string_cleanup(&str);

	++cookie;
	if (!bro_record_add_val(rec, EV_COOKIE, BRO_TYPE_COUNT,
	    NULL, &cookie)) {
		fprintf(stderr, "%s: bro_record_add_val(%s) failed\n",
		    prog, EV_CMD);
		bro_record_free(rec);
		return (NULL);
	}

	ev = bro_event_new(requestevstr);

	bro_event_add_val(ev, BRO_TYPE_RECORD, NULL, rec);
	bro_record_free(rec);
	return (ev);
}

void
callback(BroConn *bc, void *data, BroRecord *rec)
{
	uint32 cookie;
	double time;
	char cmd[64];
	char payload[1024];

	if (!extractstr(rec, EV_CMD, cmd, sizeof(cmd))) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_CMD);
		return;
	}
	if (!extractuint32(rec, EV_COOKIE, &cookie)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_COOKIE);
		return;
	}
	if (!extractbool(rec, EV_SUCCESS, &success)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_SUCCESS);
		return;
	}
	if (!extracttime(rec, EV_TIME, &time)) {
		fprintf(stderr, "%s: missing %s\n", prog, EV_TIME);
		return;
	}
	fprintf(stderr, "%s: %.6f recv:%d \"%s\"\n", prog, time, cookie, cmd);

	/* Payload goes to stdout */
	if (extractstr(rec, EV_PAYLOAD, payload, sizeof(payload))) {
		fputs(payload, stdout);
		fflush(stdout);
	}
	if (!success)
		fprintf(stderr, "%s: %s failed\n", prog, cmd);

}

int
command(int s, BroConn *bc, int an, char **av)
{
	int n, nfds;
	size_t size, len;
	char *cp;
	BroEvent *ev;
	struct timeval tmo;
	fd_set readfds;
	char buf[1024];

	size = sizeof(buf);
	cp = buf;
	for (; an > 0; --an, ++av) {
		len = strlen(*av);
		/* room for blank and eos */
		if (size < len + 2) {
			fprintf(stderr, "%s: command: too big!\n", prog);
			exit(1);
		}
		if (cp != buf) {
			*cp++ = ' ';
			--size;
		}
		strcpy(cp, *av);
		cp += len;
		size -= len;
	}
	ev = acldevent(buf);

	if (!bro_event_send(bc, ev)) {
		fprintf(stderr, "%s: failed to send\n",
		    prog);
		return (1);
	}
	bro_event_free(ev);

	for (;;) {
		/* Recalculate fdsets */
		FD_ZERO(&readfds);
		FD_SET(s, &readfds);
		nfds = s;
		++nfds;

		/* Set timeout */
		tmo.tv_sec = timeout;
		tmo.tv_usec = 0;
		n = select(nfds, &readfds, NULL, NULL, &tmo);

		if (n < 0) {
			/* Don't choke if we get ptraced */
			if (errno == EINTR)
				continue;
			fprintf(stderr, "%s: select: %s\n",
			    prog, strerror(errno));
			exit(EX_OSERR);
		}

		/* Timeout */
		if (n == 0) {
			// bro_event_queue_flush(bc);
			fprintf(stderr, "%s: timeout\n", prog);
			return (1);
		}

		/* Data from server */
		if (FD_ISSET(s, &readfds)) {
			--n;
			if (!bro_conn_process_input(bc) &&
			    !bro_conn_alive(bc)) {
				fprintf(stderr, "%s: server exited\n", prog);
				return (1);
			}
			/* Got our response; bail! */
			return (success == 0);
		}

		if (n != 0)
			fprintf(stderr, "%s: select leakage: %d\n", prog, n);
			continue;
	}
}

int
loop(int s, BroConn *bc)
{
	int n, nfds, ret;
	ssize_t cc;
	BroEvent *ev;
	struct timeval tmo;
	fd_set readfds;
	char buf[1024];
	const char *prompt = "> ";

	ret = 0;
	fputs(prompt, stdout);
	fflush(stdout);
	for (;;) {
		/* Recalculate fdsets */
		FD_ZERO(&readfds);
		FD_SET(fileno(stdin), &readfds);
		nfds = fileno(stdin);
		FD_SET(s, &readfds);
		if (nfds < s)
			nfds = s;
		++nfds;

		/* Set timeout */
		tmo.tv_sec = timeout;
		tmo.tv_usec = 0;
		n = select(nfds, &readfds, NULL, NULL, &tmo);

		if (n < 0) {
			/* Don't choke if we get ptraced */
			if (errno == EINTR)
				continue;
			fprintf(stderr, "%s: select: %s\n",
			    prog, strerror(errno));
			exit(EX_OSERR);
		}

		/* Timeout */
		if (n == 0)
			bro_event_queue_flush(bc);

		/* Data from server */
		if (FD_ISSET(s, &readfds)) {
			--n;
			if (!bro_conn_process_input(bc) &&
			    !bro_conn_alive(bc)) {
				fprintf(stderr, "%s: server exited\n", prog);
				return (1);
			}
			fputs(prompt, stdout);
			fflush(stdout);
		}

		/* Input from stdin */
		if (FD_ISSET(fileno(stdin), &readfds)) {
			--n;
			cc = read(fileno(stdin), buf, sizeof(buf));
			if (cc < 0)
				fprintf(stderr, "%s: read stdin: %s\n",
				    prog, strerror(errno));
			else if (cc == 0) {
				/* EOF */
				putc('\n', stdout);
				return (ret);
			} else {
				if (cc >= sizeof(buf))
					cc = sizeof(buf) - 1;
				if (cc > 0 && buf[cc - 1] == '\n')
					--cc;
				buf[cc] = '\0';
				ev = acldevent(buf);

				if (!bro_event_send(bc, ev)) {
					fprintf(stderr, "%s: failed to send\n",
					    prog);
					ret = 1;
				}
				bro_event_free(ev);
				if (!success)
					ret = 1;
			}
		}

		if (n != 0)
			fprintf(stderr, "%s: select leakage: %d\n", prog, n);
			continue;
	}
}

__dead void
usage(void)
{

	fprintf(stderr, "%s version %s\n", prog, version);
	fprintf(stderr, "Usage: %s [-dDv] [-a addr] [-P port] [-e requestevstr]"
	    " [-E replyevstr] [-f file] [-c config] \\\n\t {command option}\n",
		prog);
	exit(EX_USAGE);
}

/*
 * common - example broccoli common code
 */

int
extractbool(BroRecord *rec, const char *name, int *bp)
{
	int btype;
	int *bbp;

	btype = BRO_TYPE_BOOL;
	bbp = bro_record_get_named_val(rec, name, &btype);
	if (bbp == NULL) {
		*bp = 0;
		return (0);
	}
	*bp = *bbp;
	return (1);
}

int
extractstr(BroRecord *rec, const char *name, char *str, size_t size)
{
	int btype;
	BroString *bstr;

	btype = BRO_TYPE_STRING;
	bstr = bro_record_get_named_val(rec, name, &btype);
	if (bstr == NULL)
		return (0);
	if (size > bstr->str_len + 1)
		size = bstr->str_len + 1;
	strncpy(str, bstr->str_val, size);
	str[size] = '\0';
	bro_string_cleanup(bstr);
	return (1);
}

int
extracttime(BroRecord *rec, const char *name, double *dp)
{
	int btype;
	double *bdp;

	btype = BRO_TYPE_TIME;
	bdp = bro_record_get_named_val(rec, name, &btype);
	if (bdp == NULL) {
		*dp = 0;
		return (0);
	}
	*dp = *bdp;
	return (1);
}

int
extractuint32(BroRecord *rec, const char *name, uint32 *up)
{
	int btype;
	uint32 *bup;

	btype = BRO_TYPE_COUNT;
	bup = bro_record_get_named_val(rec, name, &btype);
	if (bup == NULL) {
		*up = 0;
		return (0);
	}
	*up = *bup;
	return (1);
}

void *
new(size_t num, size_t size, const char *what)
{
	void *p;

	p = calloc(num, size);
	if (p == NULL) {
		fprintf(stderr, "%s: new: calloc %s: %s\n",
		    prog, what, strerror(errno));
		exit(EX_OSERR);
	}
	return (p);
}
