/*
 * Copyright (c) 2002, 2003, 2004, 2006, 2008
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: cf.c 529 2008-10-24 06:57:49Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>

#include "acld.h"
#include "cf.h"

/* Locals */
static int errors;

/* Forwards */
static int readcf(const char *, struct cf *, int);

/* Open, read and parse one config file */
static int
readcf(const char *fn, struct cf *cf, int mustexist)
{
	int n, errs, *ip, *ip2;
	u_int32_t *uip;
	char *cp, *cp2, **pp;
	time_t t;
	FILE *f;
	struct in_addr in;
	struct stat sbuf;
	char buf[1024];
	char at[1024];

	errs = 0;
	f = fopen(fn, "r");
	if (f == NULL) {
		if (mustexist) {
			lg(LOG_ERR, "readcf: fopen %s: %s",
			    fn, strerror(errno));
			++errs;
		}
		return (errs);
	}

	/* Get the config file mod time */
	if (fstat(fileno(f), &sbuf) >= 0) {
		t = (sbuf.st_mtime > sbuf.st_ctime) ?
		    sbuf.st_mtime : sbuf.st_ctime;
		if (cf->c_time < t)
			cf->c_time = t;
	} else {
		lg(LOG_ERR, "readcf: fstat %s: %s", fn, strerror(errno));
		++errs;
	}

	n = 0;
	while (fgets(buf, sizeof(buf), f) != NULL) {
		++n;

		/* Eat comments */
		cp = strchr(buf, '#');
		if (cp)
			*cp = '\0';

		/* Eat trailing whitesapce */
		cp = buf + strlen(buf) - 1;
		while (cp >= buf && isspace((int)*cp))
			*cp-- = '\0';
		cp = buf;

		/* Skip blank lines */
		if (*cp == '\n' || *cp == '\0')
			continue;

		/* Get command */
		cp2 = cp;
		while (!isspace((int)*cp) && *cp != '\0')
			++cp;
		*cp++ = '\0';

		(void)sprintf(at, "%s:%d", fn, n);

		/* Find start of next keyword */
		while (isspace((int)*cp))
			++cp;

		/* Quoted one line commands */
		if ((pp = &cf->c_cuser, strcasecmp(cp2, "cuser") == 0) ||
		    (pp = &cf->c_cpass1, strcasecmp(cp2, "cpass1") == 0) ||
		    (pp = &cf->c_cpass2, strcasecmp(cp2, "cpass2") == 0) ||
		    (pp = &cf->c_euser, strcasecmp(cp2, "euser") == 0) ||
		    (pp = &cf->c_epass1, strcasecmp(cp2, "epass1") == 0) ||
		    (pp = &cf->c_epass2, strcasecmp(cp2, "epass2") == 0) ||
		    (pp = &cf->c_expect, strcasecmp(cp2, "expect") == 0) ||
		    (pp = &cf->c_router, strcasecmp(cp2, "router") == 0) ||
		    (pp = &cf->c_script, strcasecmp(cp2, "script") == 0) ||
		    (pp = &cf->c_whitelistfn,
			strcasecmp(cp2, "whitelist") == 0)) {
			if (*cp++ != '"') {
				lg(LOG_ERR, "%s missing leading double quote",
				    at);
				++errors;
				continue;
			}
			if (*pp != NULL)
				free(*pp);
			*pp = strsave(cp);
			cp = *pp;
			cp2 = cp + strlen(cp) - 1;
			if (cp > cp2 || *cp2 != '"') {
				lg(LOG_ERR, "%s missing trailing double quote",
				    at);
				++errors;
				free(*pp);
				*pp = NULL;
				continue;
			}
			*cp2 = '\0';
			continue;
		}

		/* Commands with numeric arguments */
		if ((ip = &cf->c_ayt_secs, strcasecmp(cp2, "ayt_secs") == 0) ||
		    (ip = &cf->c_incrseq, strcasecmp(cp2, "incrseq") == 0) ||
		    (ip = &cf->c_login_secs,
			strcasecmp(cp2, "login_secs") == 0) ||
		    (ip = &cf->c_nullzeromax,
			strcasecmp(cp2, "nullzeromax") == 0) ||
		    (ip = &cf->c_maxseq, strcasecmp(cp2, "maxseq") == 0) ||
		    (ip = &cf->c_port, strcasecmp(cp2, "port") == 0) ||
#ifdef HAVE_BROCCOLI
		    (ip = &cf->c_portbro, strcasecmp(cp2, "portbro") == 0) ||
#endif
		    (ip = &cf->c_portro, strcasecmp(cp2, "portro") == 0) ||
		    (ip = &cf->c_portweb, strcasecmp(cp2, "portweb") == 0) ||
		    (ip = &cf->c_select_secs,
			strcasecmp(cp2, "select_secs") == 0) ||
		    (ip = &cf->c_sync_secs,
			strcasecmp(cp2, "sync_secs") == 0)) {
			*ip = atoi(cp);
			if (*cp == '-' || *cp == '+')
				++cp;
			while (isdigit((int)*cp))
				++cp;
			while (isspace((int)*cp))
				++cp;
			if (*cp != '\0') {
				lg(LOG_ERR, "%s trailing garbage (%s)", at, cp);
				++errors;
			}
			continue;
		}

		/* Commands with two numeric arguments */
		if ((ip = &cf->c_lowseq, ip2 = &cf->c_highseq,
			strcasecmp(cp2, "seqrange") == 0) ||
		    (ip = &cf->c_lowportseq, ip2 = &cf->c_highportseq,
			strcasecmp(cp2, "portseqrange") == 0) ||
		    (ip = &cf->c_lowpermithostportseq,
			ip2 = &cf->c_highpermithostportseq,
			strcasecmp(cp2, "permithostportseqrange") == 0)) {
			/* Low value */
			*ip = atoi(cp);
			if (*cp == '-' || *cp == '+')
				++cp;
			if (!isdigit((int)*cp)) {
				lg(LOG_ERR, "%s missing low %s", at, cp2);
				++errors;
			}
			while (isdigit((int)*cp))
				++cp;
			if (!isspace((int)*cp)) {
				lg(LOG_ERR, "%s missing %s whitespace",
				    at, cp2);
				++errors;
			}
			while (isspace((int)*cp))
				++cp;

			/* High value */
			*ip2 = atoi(cp);
			if (*cp == '-' || *cp == '+')
				++cp;
			if (!isdigit((int)*cp)) {
				lg(LOG_ERR, "%s missing high %s", at, cp2);
				++errors;
			}
			while (isdigit((int)*cp))
				++cp;
			while (isspace((int)*cp))
				++cp;
			if (*cp != '\0') {
				lg(LOG_ERR, "%s %s trailing garbage", at, cp2);
				++errors;
			}
			continue;
		}

		/* Commands with ip address arguments */
		if ((uip = &cf->c_bindaddr, strcasecmp(cp2, "bindaddr") == 0)
#ifdef HAVE_BROCCOLI
		    || (uip = &cf->c_broaddr, strcasecmp(cp2, "broaddr") == 0)
#endif
		    ) {
			if (!inet_aton(cp, &in)) {
				lg(LOG_ERR, "%s bogus ip address (%s)", at, cp);
				++errors;
			}
			*uip = in.s_addr;
			continue;
		}

		/* ACLs */
		if (strcasecmp(cp2, "acl") == 0) {
			acladdacllist(cf, cp);
			continue;
		}

		/* Interfaces */
		if (strcasecmp(cp2, "interface") == 0) {
			intlistadd(cf, cp);
			continue;
		}

		/* Null zero nets */
		if (strcasecmp(cp2, "nullzeronet") == 0) {
			if (!nullzeronetadd(cf, cp))
				++errors;
			continue;
		}

		/* nets_log() syslog facility */
		if (strcasecmp(cp2, "netsfac") == 0) {
			cf->c_netsfac = str2val(syslog2str, cp);
			if (cf->c_netsfac < 0) {
				lg(LOG_ERR, "error: bogus netpri value \"%s\"",
				    cp);
				++errors;
			}
			continue;
		}

		/* Bad commands */
		lg(LOG_ERR, "%s unknown command \"%s\"", at, cp2);
		++errors;
	}
	(void)fclose(f);

	return (errs);
}

struct cf *
parsecf(const char *fn)
{
	int i, j, seqrange, portseqrange, permithostportseqrange;
	char *cp;
	struct cf *cf;
	struct acllist *ap;
	struct intlist *ip;

	cf = (struct cf *)new(1, sizeof(*cf), "parsecf cf");

	/* Defaults */
	cf->c_ayt_secs = AYT_SECS;
	cf->c_login_secs = LOGIN_SECS;
	cf->c_select_secs = SELECT_SECS;
	cf->c_sync_secs = SYNC_SECS;
	cf->c_incrseq = 1;
	cf->c_bindaddr = inet_addr("127.0.0.1");
#ifdef HAVE_BROCCOLI
	cf->c_broaddr = ~0;		/* flag so we can tell if it was set */
	cf->c_portbro = 0;
#endif

	errors += readcf(fn, cf, 1);

	if ((cp = "router", cf->c_router == NULL) ||
	    (cp = "script", cf->c_script == NULL) ||
	    (cp = "expect", cf->c_expect == NULL) ||
	    (cp = "acl", cf->c_acllist == NULL) ||
	    (cp = "interface", cf->c_intlist == NULL)) {
		lg(LOG_ERR, "error: \"%s\" missing but required", cp);
		++errors;
	}

	if ((cp = "login_secs", cf->c_login_secs <= 0) ||
	    (cp = "port", cf->c_port <= 0) ||
	    (cp = "select_secs", cf->c_select_secs <= 0)) {
		lg(LOG_ERR, "error: \"%s\" must be non-zero", cp);
		++errors;
	}

	if ((cf->c_portro > 0 && (cp = "portro", cf->c_portro == cf->c_port)) ||
	    (cf->c_portweb > 0 && (cp = "portweb",
		cf->c_portweb == cf->c_port))) {
		lg(LOG_ERR, "error: \"%s\" can't be the same as \"port\"", cp);
		++errors;
	}
	if (cf->c_portro > 0 && cf->c_portweb > 0 &&
	    cf->c_portro == cf->c_portweb) {
		lg(LOG_ERR,
		    "error: \"portro\" can't be the same as \"portweb\"");
		++errors;
	}

	seqrange = (cf->c_lowseq != 0 || cf->c_highseq != 0);
	portseqrange = (cf->c_lowportseq != 0 || cf->c_highportseq != 0);
	permithostportseqrange = (cf->c_lowpermithostportseq != 0 ||
	    cf->c_highpermithostportseq != 0);

#define IN_SEQRANGE(n) ((n) >= cf->c_lowseq && (n) <= cf->c_highseq)
#define IN_PORTSEQRANGE(n) ((n) >= cf->c_lowportseq && (n) <= cf->c_highportseq)
#define IN_PERMITHOSTPORTSEQRANGE(n) \
    ((n) >= cf->c_lowpermithostportseq && (n) <= cf->c_highpermithostportseq)

	if (seqrange) {
		if (cf->c_lowseq >= cf->c_highseq) {
			lg(LOG_ERR, "error: \"seqrange\" must be increasing");
			++errors;
		}
		if (portseqrange && (IN_SEQRANGE(cf->c_lowportseq) ||
		    IN_SEQRANGE(cf->c_highportseq))) {
			lg(LOG_ERR,
			    "error: \"seqrange\" and \"portseqrange\" overlap");
			++errors;
		}
	}

	if (portseqrange) {
		if (cf->c_lowportseq >= cf->c_highportseq) {
			lg(LOG_ERR,
			    "error: \"portseqrange\" must be increasing");
			++errors;
		}
		if (permithostportseqrange &&
		    (IN_PERMITHOSTPORTSEQRANGE(cf->c_lowseq) ||
		    IN_PERMITHOSTPORTSEQRANGE(cf->c_highseq))) {
			lg(LOG_ERR, "error: \"portseqrange\" and"
			    " \"permithostportseqrange\" overlap");
			++errors;
		}
	}

	if (permithostportseqrange) {
		if (cf->c_lowpermithostportseq >= cf->c_highpermithostportseq) {
			lg(LOG_ERR, "error: \"permithostportseqrange\" must"
			" be increasing");
			++errors;
		}
		if (seqrange &&
		    (IN_SEQRANGE(cf->c_lowpermithostportseq) ||
		    IN_SEQRANGE(cf->c_highpermithostportseq))) {
			lg(LOG_ERR, "error: \"seqrange\" and"
			    " \"permithostportseqrange\" overlap");
			++errors;
		}
	}
#define GOOD_PORT(p) ((p) > 0 && (p) <= 2 << 16)
	if (cf->c_port != 0 && !GOOD_PORT(cf->c_port)) {
		lg(LOG_ERR, "error: bad \"port\" port value");
		++errors;
	}
#ifdef HAVE_BROCCOLI
	if (cf->c_portbro != 0 && !GOOD_PORT(cf->c_portbro)) {
		lg(LOG_ERR, "error: bad \"portbro\" port value");
		++errors;
	}
#endif
	if (cf->c_portro != 0 && !GOOD_PORT(cf->c_portro)) {
		lg(LOG_ERR, "error: bad \"portro\" port value");
		++errors;
	}
	if (cf->c_portweb != 0 && !GOOD_PORT(cf->c_portweb)) {
		lg(LOG_ERR, "error: bad \"portweb\" port value");
		++errors;
	}

#define ESCAPESTR(str) { \
    cp = str; \
    str = strsave(escapestr(cp)); \
    if (cp != NULL) \
	    free(cp); \
    }

	ESCAPESTR(cf->c_cuser);
	ESCAPESTR(cf->c_cpass1);
	ESCAPESTR(cf->c_cpass2);
	ESCAPESTR(cf->c_euser);
	ESCAPESTR(cf->c_epass1);
	ESCAPESTR(cf->c_epass2);

	/* Hook interfaces into acls */
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		for (j = 0, ip = cf->c_intlist; j < cf->c_intlistlen; ++j, ++ip)
			if (strcmp(ap->name, ip->i_acl) == 0) {
				ap->intlist = ip;
				break;
			}
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		if (ap->intlist == NULL) {
			lg(LOG_ERR, "error: ACL %s missing interface",
			    ap->name);
			++errors;
		}

	/* Whitelist */
	whitelistread(cf);

#ifdef HAVE_BROCCOLI
	/* Default broaddr to bindaddr */
	if (cf->c_broaddr == ~0)
		cf->c_broaddr = cf->c_bindaddr;
#endif

	if (errors > 0)
		exit(EX_CONFIG);

	return (cf);
}

static void
freeone(void *p)
{

	if (p != NULL)
		free(p);
}

void
freecf(struct cf *cf)
{
	int i;
	struct acllist *ap;
	struct intlist *ip;

	/* Single line options */
	freeone(cf->c_cpass1);
	freeone(cf->c_cpass2);
	freeone(cf->c_epass1);
	freeone(cf->c_epass2);
	freeone(cf->c_expect);
	freeone(cf->c_router);
	freeone(cf->c_script);

	/* ACLs */
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		aclfree(ap);
	cf->c_acllist = NULL;
	cf->c_acllistlen = 0;
	cf->c_acllistsize = 0;

	/* Interfaces */
	for (i = 0, ip = cf->c_intlist; i < cf->c_intlistlen; ++i, ++ip)
		intlistfree(ip);
	cf->c_intlist = NULL;
	cf->c_intlistlen = 0;
	cf->c_intlistsize = 0;

	/* Null zero routes */
	if (cf->c_nullzeronets != NULL) {
		free(cf->c_nullzeronets);
		cf->c_nullzeronets = NULL;
		cf->c_nullzeronetslen = 0;
		cf->c_nullzeronetssize = 0;
	}

	/* Whitelist */
	if (cf->c_whitelist != NULL) {
		free(cf->c_whitelist);
		cf->c_whitelist = NULL;
		cf->c_whitelistlen = 0;
		cf->c_whitelistsize = 0;
	}
	if (cf->c_whitelistfn != NULL) {
		free(cf->c_whitelistfn);
		cf->c_whitelistfn = NULL;
	}

	free(cf);
}
